/*
 * Copyright (C) 2000-2004 the xine project
 * 
 * This file is part of xine, a free video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * Time stretch by a given factor, optionally preserving pitch
 *
 * $Id: stretch.c,v 1.8 2006/01/27 07:46:12 tmattern Exp $
 *
 */
 
/*
 *  Modifications for use in VC/DC Copyright (C) 2008 Josh Sutherland <OneThirty8@aol.com>
 * Preservation of pitch here is not optional--that's the desired effect!
 */


#include <stdio.h>
#include "config.h"

#define INT16_MIN -32768
#define INT16_MAX 32767
#define CLIP_INT16(s) ((s) < INT16_MIN) ? INT16_MIN : \
                      (((s) > INT16_MAX) ? INT16_MAX : (s))



static void stretch_process_fragment(int16_t* input_samples, uint32_t in_samples, 
				int16_t* output_samples, uint32_t out_samples, int channels)
{

  int num_frames_in = in_samples;
  int num_frames_out = out_samples;

     if( num_frames_in > num_frames_out )
     {
       /*
        * time compressing strategy
        *
        * input chunk has two halves, A and B.
        * output chunk is composed as follow:
        * - some frames copied directly from A
        * - some frames copied from A merged with frames from B 
        *   weighted by an increasing factor (0 -> 1.0)
        * - frames from A weighted by a decreasing factor (1.0 -> 0)
        *   merged with frames copied from B
        * - some frames copied directly from B
        */
        
       int merge_frames = num_frames_in - num_frames_out;
       int copy_frames;
       int16_t *src = input_samples;
       int16_t *dst = output_samples;
       int i, j;
       double w;
       if( merge_frames > num_frames_out )
         merge_frames = num_frames_out;
       copy_frames = num_frames_out - merge_frames;
       
       w = 1.0f / (merge_frames/2);
       memcpy(dst, src, copy_frames/2 * (2*channels));
       dst += copy_frames/2 * channels;
       src += copy_frames/2 * channels;
 
       for( i = 0; i < merge_frames/2; i++ )
       {
         for( j = 0; j < channels; j++, src++, dst++ ) {
                                          
           int32_t  s = (int32_t) ((double) src[0] +
                                    src[merge_frames * channels] * (w*i));
           *dst = CLIP_INT16(s);
         }
       }
       
       for( ; i < merge_frames; i++ )
       {
         for( j = 0; j < channels; j++, src++, dst++ ) {
                                          
           int32_t  s = (int32_t) ((double) src[0] * (w*i) +
                                    src[merge_frames * channels]);
           *dst = CLIP_INT16(s);
         }
       }
       
       src += merge_frames * channels;
       
       memcpy(dst, src, (copy_frames - copy_frames/2) * (2*channels));
     
     } else {
       /*
        * time expansion strategy
        *
        * output chunk is composed of two versions of the
        * input chunk:
        * - first part copied directly from input, and then
        *   merged with the second (delayed) part using a 
        *   decreasing factor (1.0 -> 0)
        * - the delayed version of the input is merged with
        *   an increasing factor (0 -> 1.0) and then (when 
        *   factor reaches 1.0) just copied until the end. 
        */
       
       int merge_frames = num_frames_out - num_frames_in;
       int copy_frames = num_frames_out - merge_frames;
       int16_t *src1 = input_samples;
       int16_t *src2;
       int16_t *dst = output_samples;
       int i, j;
       double w;
       
       memcpy(dst, src1, copy_frames/2 * (2*channels));
       dst += copy_frames/2 * channels;
       src1 += copy_frames/2 * channels;
       src2 = src1 - merge_frames * channels;
       w = 1.0f / (merge_frames/2);
       
       for( i = 0; i < merge_frames/2; i++ )
       {
         for( j = 0; j < channels; j++, src1++, src2++, dst++ ) {
                                          
           int32_t  s = (int32_t) ((double) *src1 +
                                    *src2 * (w*i));
           *dst = CLIP_INT16(s);
         }
       }
       
       for( ; i < merge_frames; i++ )
       {
         for( j = 0; j < channels; j++, src1++, src2++, dst++ ) {
                                          
           int32_t  s = (int32_t) ((double) *src1 * (w*i) +
                                    *src2);
           *dst = CLIP_INT16(s);
         }
       }
              
       memcpy(dst, src2, (copy_frames - copy_frames/2) *
                         (2*channels));

    }
      
  /* copy processed fragment into multiple audio buffers, if needed */
  /* while( num_frames_out ) {
    outbuf = port->original_port->get_buffer(port->original_port); 
  
    outbuf->num_frames = outbuf->mem_size / this->bytes_per_frame;
    if( outbuf->num_frames > num_frames_out )
      outbuf->num_frames = num_frames_out;
        
    memcpy( outbuf->mem, data_out, 
            outbuf->num_frames * this->bytes_per_frame );
    num_frames_out -= outbuf->num_frames;
    data_out = (uint16_t *)((uint8_t *)data_out + outbuf->num_frames * this->bytes_per_frame);
                
    outbuf->vpts        = this->pts;
    this->pts           = 0;
    outbuf->stream      = stream;
    outbuf->format.bits = port->bits;
    outbuf->format.rate = port->rate;
    outbuf->format.mode = port->mode;
    
    _x_extra_info_merge( outbuf->extra_info, extra_info );
    
    port->original_port->put_buffer(port->original_port, outbuf, stream );  
  }
  
  this->num_frames = 0; */
}

void stretch_mono(int16_t* input_samples, uint32_t in_samples, 
				int16_t* output_samples, uint32_t out_samples)
{
  stretch_process_fragment(input_samples, in_samples, output_samples, out_samples, 1);
}
void stretch_stereo(int16_t* input_samples, uint32_t in_samples, 
				int16_t* output_samples, uint32_t out_samples)
{
  stretch_process_fragment(input_samples, in_samples, output_samples, out_samples, 2);
}
