/*
 * vidclock.c
 * Copyright (C) 2008 Josh Sutherland <OneThirty8@aol.com>
 * This file is part of VC/DC, a free MPEG-2 video player.
 *
 * VC/DC is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * VC/DC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * This is based on an idea I got from xine-lib, looking at code in
 * xine-engine/metronom.c and xine-engine/resample.c
 * Aditionally, much code is taken from or modeled after mpeg2dec by Michel Lespinasse
 * and Aaron Holtzman, particularly the print_fps (int final); function, which
 * is reused here to check whether we're running fast or slow.
 */
  
#include <kos.h>

#include "config.h"
#include "inttypes.h"
#include "mpeg2.h"
#include "gettimeofday.h"

#define difference(a, b) ((a > b) ? a-b : b-a)

extern int seeking; /* Make this private to the vidclock struct with functions to set and get. */

static struct vidclock_s{
  uint32_t frame_counter;
  uint32_t last_count;
  double fps;
  double cur_fps; /* We'll keep track of this for about 1/8 of a second to decide whether or
  			not to drop video frames. */
  double ptime_inc; /* Useful for supporting ptime in VCD PlayLists as well as a/v synch. */
  
  double ptime_total; /* How many ptime units have passed since starting playback. */
  
  int incoming_reset, decoded_reset; /* Set these to 1 if we just started/skipped/whatever. */
  int skipped;  /* Useful if debugging--throw some code in somewhere to see how many frames you're skipping. */
  /* struct timeval recent_frames[6]; */ /* Here we keep track of the framerate over six frames.  If we skip a
  						frame, we need to update it anyway so the video output thread knows
						how to fake the 'real' framerate. */
  /* Framerate regulation stuff */
  uint64_t start_time;
  int frame_period; /* Really only useful for debugging... */

  uint64_t last_frame_time; /* The time it was last time we got a frame. Useful for closer a/v synch? */
  int audio_rate;  /* The size, in samples, of a frame of audio. */
  int sample_counter;
  int audio_latency;
  int max_ahead;  /* How much ahead or behind we'll allow the audio to drift before bothering to correct. */
  int max_behind;
} vidclock;

static mutex_t vidclock_mut = MUTEX_INITIALIZER;

void vidclock_init(unsigned int frame_period)
{
  vidclock.frame_counter = vidclock.last_count = 0;
  
  /* Frame period is stored assuming a 27 MHz clock.
  Dividing 27 million by the frame period will give us the
  frame rate. */
  vidclock.fps = 27000000.0 / frame_period;
  vidclock.cur_fps = vidclock.fps;
  
 /* Calculate the ptime value for this sequence. */
   vidclock.ptime_total = 0.0f;
   vidclock.ptime_inc = (1.0f/vidclock.fps);
   
  vidclock.incoming_reset = vidclock.decoded_reset = 1;
  vidclock.audio_rate = 0;  /* If we just started, we have no clue what this is. */
  vidclock.sample_counter = 0;
  vidclock.last_frame_time = timer_ms_gettime64();
#ifdef DEBUGFPS
clear_notes();
#endif  
}

/* This is good enough for video-only sequences, but for streams with audio we can
   probably be even more accurate by looking at the number of samples that have played. */
int vidclock_get_ptime()
{
  double chunks;
  chunks = (vidclock.ptime_total * 15.0f);
  return (uint16)chunks;
}


void vidclock_reset()
{
  vidclock.cur_fps = vidclock.fps;
  vidclock.last_count = vidclock.frame_counter =
  vidclock.audio_rate = vidclock.sample_counter = 0;
  
  vidclock.incoming_reset = vidclock.decoded_reset = 1;
}

void vidclock_set_audio_rate(int freqs, int chans, int buf_size)
{
  vidclock.audio_rate = freqs;
  
  /* Account for latency in samples by dividing the buffer size by 2
     times the number of channels. */ 
  vidclock.audio_latency = buf_size/(chans*2);
  
  /* One full frame (PAL) might be a bit much. */
  vidclock.max_ahead = freqs/50;//freqs/25;
  vidclock.max_behind = 0 - vidclock.max_ahead;  
}

#if 0
int vidclock_drift_allowance()
{
 if(vidclock.audio_rate && vidclock.fps)
  return(int)(vidclock.audio_rate/(vidclock.fps*2)); /* Basically, half a frame isn't too horribly out of synch. */
 else
  return 0;
}
#endif

/* We should probably add a check for out-of-synch and have a threshold to go by--that way, if we're close enough,
   just go with what we have and reset counters and whatnot below. */
int vidclock_handle_audio(uint32_t samples)
{
  double vidsecs; /* How much time, in seconds, has passed since we started/restarted. */
  uint64_t now, retval;
  int vidsecs_to_samples; /* Given how much time has passed, how many samples we should have. */
  static int last_framenum = 0; /* Use this to check that we don't resample too often! */
  static int cum_drift = 0; /* Add up the out of synch values each call, and average every 10th. */
  static int synch_counter = 0; /* Keep track of the number of calls. */
  int current_drift = 0; /* How much we have drifted as of right this moment. */ 
  
  vidclock.sample_counter += samples;
   
   if (vidclock.frame_counter == 0)
   {
     last_framenum = 0;
     vidclock.sample_counter = samples;
     vidclock.last_frame_time = timer_ms_gettime64(); /* Not perfect, but it will drift back quickly. */
     synch_counter = 0;
     cum_drift = 0;
     return 0;
   }
   if (vidclock.decoded_reset==1){
	vidclock.start_time = timer_ms_gettime64(); 
	vidclock.decoded_reset = 0;
   }
   //if ((vidclock.frame_counter - last_framenum) * vidclock.ptime_inc < 1.0f/*0.5*/) return 0; /* So we don't resample too often. */
   
   //last_framenum = vidclock.frame_counter;
   
   now = timer_ms_gettime64();
   vidsecs = (now - vidclock.start_time) / 1000.0f;
   
   vidsecs_to_samples = (int)(vidsecs * vidclock.audio_rate);
   
   /* A positive value means that the audio is behind, negative that it's
   ahead. */
   current_drift = vidsecs_to_samples - (vidclock.sample_counter - vidclock.audio_latency);
   //cum_drift += vidsecs_to_samples - (vidclock.sample_counter - vidclock.audio_latency);
   
   if (vidclock.max_behind < current_drift < vidclock.max_ahead) /* We're in synch! */
   {
     synch_counter = 0;
     cum_drift = 0;
     return 0;
   }
   
   cum_drift += current_drift;
   synch_counter++;
   if (synch_counter < 10) return 0;
   
   retval = cum_drift/synch_counter;//10;
   cum_drift = synch_counter = 0;
   return retval;
   //return vidsecs_to_samples - (vidclock.sample_counter - vidclock.audio_latency);   
}

/* This could probably be handled better.  We should find out if we're behind, not if we're running slow over the last
  few frames.  We'll probably skip fewer frames on stuff that isn't huge in display size or bitrate, and probably
  be better at dropping enough frames on other stuff. */
void vidclock_handle_incoming (const mpeg2_info_t * info, mpeg2dec_t * mpeg2dec)
{
    static uint32_t frame_counter = 0;
    static struct timeval tv_beg[6], tv_start;  /* Having 6 tv_beg timevalues will allow us to keep track of
    						the framerate for ~ 1/4 second (NTSC-Film).  This should be enough? */
    static int total_elapsed;
    static int last_count = 0;
    struct timeval tv_end;
    static double fps, tfps;
    int frames, elapsed;
    int i;

    gettimeofday (&tv_end, NULL);

    if (vidclock.incoming_reset==1){
        frame_counter = 0; vidclock.cur_fps = vidclock.fps; vidclock.skipped = 0;
	tv_start = tv_beg[0] = tv_beg[1] = tv_beg[2] = tv_beg[3] = tv_beg[4] = tv_beg[5] =
	tv_end;
	vidclock.incoming_reset = 0;
    }
    

    elapsed = (tv_end.tv_sec - tv_beg[0].tv_sec) * 100 +
	(tv_end.tv_usec - tv_beg[0].tv_usec) / 10000;
    total_elapsed = (tv_end.tv_sec - tv_start.tv_sec) * 100 +
	(tv_end.tv_usec - tv_start.tv_usec) / 10000;


    frame_counter++;

    for ( i = 0; i < 5; i++)
    {
      tv_beg[i] = tv_beg[i+1];
    } 
    
    tv_beg[5] = tv_end;
   
    frames = ( frame_counter < 6) ? frame_counter : 6;

    fps = frames * 100.0 / elapsed;
    tfps = frame_counter * 100.0 / total_elapsed;
#ifdef DEBUGFPS  
    take_notes("decoded framerate display", "Decoded: t:%.2f/c:%.2f", tfps, fps);
#endif    		
    last_count = frame_counter;

    /* skip B-frames if behind */ 
    if (
       ( ( (info->current_picture->flags &  PIC_MASK_CODING_TYPE) == PIC_FLAG_CODING_TYPE_B) &&
                                                                         ( (fps/vidclock.fps < 1.0) && (!(tfps/vidclock.fps > 1.0)) ) )
       || /* or skip all but every nth I-frame if we are fast-forwarding. */
       ( (seeking==1) && ( ! ( (info->current_picture->flags &  PIC_MASK_CODING_TYPE) == PIC_FLAG_CODING_TYPE_I) ) )
       )
    {
      mpeg2_skip(mpeg2dec, 1);

     vidclock.skipped ++;
            
    }else{
      mpeg2_skip(mpeg2dec, 0);
    }
    

#ifdef DEBUGFPS
     take_notes("fps skipcount","%d of %d skipped = %.2f", vidclock.skipped, frame_counter, ((double)vidclock.skipped/(double)frame_counter)) ;
#endif
   
}

/* This should be a good-enough way of regulating framerate.  If we're behind, it'll dipslay the frame immediately and if not,
  it'll wait until the time is right to display it. */
     
void vidclock_skip_frame()
{

   if (seeking) return;
   
   if (vidclock.decoded_reset==1){
	vidclock.start_time = timer_ms_gettime64(); 
	vidclock.decoded_reset = 0;
   }
   mutex_lock(&vidclock_mut);
   
   vidclock.frame_counter++;
   vidclock.ptime_total += vidclock.ptime_inc;
   vidclock.last_frame_time = timer_ms_gettime64();
   if (vidclock.decoded_reset==1)
   	vidclock.start_time = timer_ms_gettime64(); 
   mutex_unlock(&vidclock_mut);
}
/* We should be able to help out our audio synch code here by grabbing the number of samples played at
   display time and store that somewhere. */
void vidclock_handle_frame ()
{
    uint64_t now;
 
    if (vidclock.decoded_reset==1){
	vidclock.start_time = timer_ms_gettime64(); 
	vidclock.decoded_reset = 0;
    }
  mutex_lock(&vidclock_mut);

    vidclock.ptime_total += vidclock.ptime_inc;
    vidclock.frame_counter++;

 vc_wait:
    now = timer_ms_gettime64();
    if ( (now - vidclock.start_time)/1000.0f < vidclock.frame_counter * vidclock.ptime_inc)
    {
      thd_pass();
      goto vc_wait;
    }
    vidclock.cur_fps = vidclock.frame_counter / ((now - vidclock.start_time) / 1000.0f);
    mutex_unlock(&vidclock_mut);
#ifdef DEBUGFPS   
    take_notes("simulated framerate display", "Simulated framerate: %.2f /Should be:%.2f", vidclock.cur_fps, vidclock.fps);    
#endif 		
    vidclock.last_frame_time = timer_ms_gettime64();
}
